﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace SOM
{
    class SOMcs
    {
        class Row
        {
            public List<float> m_data = new List<float>(0);
            public List<int> m_pos = new List<int>(0);
        }
        List<Row> m_rows = new List<Row>(0);
        List<Row> m_map = new List<Row>(0);
        const float m_delta = 0.2f;
        const float m_decay = 0.7f;
        const int m_radius = 10;
        int[] m_mapIndex = null;
        int m_mapIndexSize = -1;
        const int m_admissibleMismatch = 5;

        private void adjustMapRow(int n, float[] vector, int size, float delta_decay)
        {
            float fMax = 0.0f;
            for (int i = 0; i < m_map[n].m_data.Count; ++i)
            {
                m_map[n].m_data[i] += delta_decay * (vector[i] - m_map[n].m_data[i]);
                if (m_map[n].m_data[i] > fMax) fMax = m_map[n].m_data[i];
            }
            for (int i = 0; i < m_map[n].m_data.Count; ++i)
            {
                m_map[n].m_data[i] /= fMax;
            }
        }

        private int makeSOMstep()
        {
            int size = m_map[0].m_data.Count;
            float[] docVector = new float[size];
            int cnt = 0;
            int nMismatch = 0;

            foreach (Row row in m_rows)
            {
                int nR = getBestMapRow(row);
                for (int i = 0; i < size; ++i)
                {
                    docVector[i] = 0.0f;
                }
                for (int i = 0; i < row.m_pos.Count; ++i)
                {
                    docVector[row.m_pos[i]] = row.m_data[i];
                }

                adjustMapRow(nR, docVector, size, m_delta);
                float delta_decay = m_delta;
                for (int i = 1; i <= m_radius; ++i)
                {
                    if ((nR + i) < m_rows.Count)
                    {
                        delta_decay *= m_decay;
                        adjustMapRow(nR + i, docVector, size, delta_decay);
                    }
                }

                delta_decay = m_delta;
                for (int i = 1; i <= m_radius; ++i)
                {
                    if ((nR - i) >= 0)
                    {
                        delta_decay *= m_decay;
                        adjustMapRow(nR - i, docVector, size, delta_decay);
                    }
                }

                if (m_mapIndex != null)
                {
                    if (m_mapIndex[cnt] != nR)
                    {
                        ++nMismatch;
                        m_mapIndex[cnt] = nR;
                    }
                    ++cnt;
                }
            }
            return nMismatch;
        }

        private int getBestCategory(int[] files)
        {
            int[] cat = new int[] { 0, 0, 0, 0, 0, 0, 0, 0 };
            for (int i = 0; i < 16; ++i)
            {
                if (files[i] < 16) ++cat[0];
                else if (files[i] < 32) ++cat[1];
                else if (files[i] < 48) ++cat[2];
                else if (files[i] < 64) ++cat[3];
                else if (files[i] < 80) ++cat[4];
                else if (files[i] < 96) ++cat[5];
                else if (files[i] < 112) ++cat[6];
                else ++cat[7];
            }

            int max = -1;
            for (int i = 0; i < 8; ++i)
            {
                if (cat[i] > max) max = cat[i];
            }

            return max;
        }

        private int getMaxCounts(int[] counts)
        {
            int max = 0;
            for (int i = 0; i < m_mapIndexSize - 16; ++i)
            {
                int cnt = 0;
                for (int j = i; j < i + 16; ++j)
                {
                    cnt += counts[j];
                }
                if (cnt > max) max = cnt;
            }
            return max;
        }

        public void estimateAccuracyForPatentCorpus128()
        {
            int[] counts = new int[m_mapIndexSize];
            int totalCount = 0;
            for (int k = 0; k < 8; ++k)
            {
                for (int i = 0; i < m_mapIndexSize; ++i) counts[i] = 0;
                for (int i = k * 16; i < (k + 1) * 16; ++i)
                {
                    ++counts[m_mapIndex[i]];
                }

                totalCount += getMaxCounts(counts);
            }
            Console.WriteLine("Accuracy = {0}", (double)(totalCount) / 128.0);
        }

        public void makeSOM()
        {
            int n = -1;
            do
            {
                n = makeSOMstep();
                Console.Write("Number of different map locations compared to previous step: {0}  \r", n);
            } while (n > m_admissibleMismatch);
            Console.WriteLine();

            Console.WriteLine("The computed map values\n");
            int cnt = 0;
            for (int i = 0; i < m_mapIndexSize; ++i)
            {
                Console.Write("{0,4}", m_mapIndex[i]);
                if (++cnt >= 16)
                {
                    Console.WriteLine();
                    cnt = 0;
                }
            }
            Console.WriteLine("\n");
        }

        private int getBestMapRow(Row row)
        {
            int nIndex = 0;
            int nMaxIndex = 0;
            float fMaxValue = 0.0f;
            foreach (Row mapRow in m_map)
            {
                float fProduct = 0.0f;
                for (int j = 0; j < row.m_pos.Count; ++j)
                {
                    int k = row.m_pos[j];
                    fProduct += row.m_data[j] * mapRow.m_data[k];
                }
                if (fProduct > fMaxValue)
                {
                    fMaxValue = fProduct;
                    nMaxIndex = nIndex;
                }
                ++nIndex;
            }
            return nMaxIndex;
        }

        public void readDataAndMakeRandomMap(string fileName)
        {
            FileInfo fi = new FileInfo(fileName);
            long nBytes = fi.Length;
            int nNonZeros = (int)(nBytes / 10);

            //read data
            int maxPos = 0;
            int nR = 0;
            int nC = 0;
            short sum = 0;
            int current_row = 0;
            m_rows.Add(new Row());
            using (FileStream stream = new FileStream(fileName, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        int nRnew = reader.ReadInt32();
                        nC = reader.ReadInt32();
                        if (nC > maxPos) maxPos = nC;
                        sum = reader.ReadInt16();
                        if (nRnew != nR)
                        {
                            m_rows.Add(new Row());
                            ++current_row;
                        }
                        nR = nRnew;
                        m_rows[current_row].m_pos.Add(nC);
                        m_rows[current_row].m_data.Add(sum);
                    }
                    reader.Close();
                }
                stream.Close();
            }
            //end

            //normalize data
            foreach (Row row in m_rows)
            {
                float max = 1.0f;
                foreach (float f in row.m_data)
                {
                    if (f > max) max = f;
                }

                for (int i = 0; i < row.m_data.Count; ++i)
                {
                    row.m_data[i] /= max;
                }
            }
            //end

            //make random  map
            Random rnd = new Random();
            int size = m_rows.Count;
            for (int i = 0; i < size; ++i)
            {
                Row row = new Row();
                for (int j = 0; j <= maxPos; ++j)
                {
                    row.m_pos.Add(j);
                    int r = rnd.Next() % 100;
                    float f = (float)r;
                    f /= 100.0f;
                    row.m_data.Add(f);
                }
                m_map.Add(row);
            }
            //end

            m_mapIndexSize = size;
            m_mapIndex = new int[size];
        }
    }
}
